#pragma once

#include <stdint.h>
#include <stdlib.h>

/* Max value for any buffer, to not squash the stack. */
#define BUFFER_MAX 4096

#ifndef __GNUC__
#define inline
#endif

/* Macros */

#define BRACKETS_THAT(that) (that)

#ifndef MAX
#define MAX(a, b) ((a) < (b) ? (b) : (a))
#endif /* MAX */

#ifndef MIN
#define MIN(a, b) ((a) > (b) ? (b) : (a))
#endif /* MIN */

#define RETURN_IF(cond, msg, ret)       \
    if (cond) {                         \
        LOG_DEBUG("DEBUG", "%s", msg);  \
        return ret;                     \
    }
#define GOTO_IF(cond, msg, label)       \
    if (cond) {                         \
        LOG_DEBUG("DEBUG", "%s", msg);  \
        goto label;                     \
    }

#define GOTO_UNLESS(cond, msg, label)   GOTO_IF(!(cond), msg, label)
#define RETURN_UNLESS(cond, msg, ret)   RETURN_IF(!(cond), msg, ret)
#define NOTHING                         /* Usefull to return nothing. */

#define STRTOL(ret, str, endptr, err_flag)                  \
{                                                           \
    err_flag = 0;                                           \
    errno    = 0;                                           \
    ret      = str == NULL ? 0 : strtol(str, &(endptr), 0); \
    err_flag = errno != 0 || endptr == str;                 \
}

extern int log_level;
enum log_level {
    ERROR   = 1,
    WARN    = 2,
    INFO    = 3,
    DEBUG   = 4,
};

#define LOG(format, level, section, ...)                                                \
    {                                                                                   \
        if (level <= log_level)                                                         \
            fprintf(stderr, " %c %s%s: " format "\n",                                   \
                level == ERROR ? '!' : level == WARN ? '*' : level == INFO ? '.' : ' ', \
                sizeof(section) > sizeof("") ? ("[" section "] ") : "",                 \
                __func__,                                                               \
                __VA_ARGS__);                                                           \
    }
#define LOG_INFO(section, format, ...)  LOG(format, INFO, section, __VA_ARGS__)
#define LOG_WARN(section, format, ...)  LOG(format, WARN, section, __VA_ARGS__)
#define LOG_ERROR(section, format, ...) LOG(format, ERROR, section, __VA_ARGS__)
#define LOG_DEBUG(section, format, ...) LOG(format, DEBUG, section, __VA_ARGS__)

/* Defines for lektor */

#define INI_MAX_LINE_LEN    512
#define INI_MAX_SECTION_LEN 80

#define URL_MAX_LEN         1024

#define SELF_EXECUTABLE_LINUX       "/proc/self/exe"
#define SELF_EXECUTABLE_FREEBSD     "/proc/curproc/file"
#define SELF_EXECUTABLE_SOLARIS     "/proc/self/path/a.out"

#define LKT_ENV_RESTART             "__LKT_RESTART"
#define LKT_ENV_CURRENT             "__LKT_CURRENT"

#define LKT_MAX_SQLITE_STATEMENT    1024
#define PROTECTED_DATABASE          "disk"

#define LKT_DB_ID                   "id"
#define LKT_DB_NAME                 "source_name"
#define LKT_DB_TITLE                "song_title"
#define LKT_DB_CAT                  "category"
#define LKT_DB_TYPE                 "song_type"
#define LKT_DB_AUTHOR               "author_name"
#define LKT_DB_LANG                 "language"
#define LKT_DB_ALL                  "string"

#define LEKTOR_TAG_MAX              256
#define LKT_MESSAGE_ARGS_MAX        32
#define LKT_MESSAGE_MAX             2048
#define LKT_DEFAULT_LIST_SIZE       10

typedef volatile enum {
    MPD_IDLE_NONE               = 0,

    MPD_IDLE_DATABASE           = ( 1 << 1  ),
    MPD_IDLE_UPDATE             = ( 1 << 2  ),
    MPD_IDLE_STORED_PLAYLIST    = ( 1 << 3  ),
    MPD_IDLE_PLAYLIST           = ( 1 << 4  ),
    MPD_IDLE_PLAYER             = ( 1 << 5  ),
    MPD_IDLE_MIXER              = ( 1 << 6  ),
    MPD_IDLE_OUTPUT             = ( 1 << 7  ),
    MPD_IDLE_OPTIONS            = ( 1 << 8  ),
    MPD_IDLE_PARTITION          = ( 1 << 9  ),
    MPD_IDLE_STICKER            = ( 1 << 10 ),
    MPD_IDLE_SUBSCRIPTION       = ( 1 << 11 ),
    MPD_IDLE_MESSAGE            = ( 1 << 12 ),

    MPD_IDLE_ALL =
        MPD_IDLE_DATABASE | MPD_IDLE_UPDATE | MPD_IDLE_STORED_PLAYLIST  |
        MPD_IDLE_PLAYLIST | MPD_IDLE_PLAYER | MPD_IDLE_MIXER            |
        MPD_IDLE_OUTPUT | MPD_IDLE_OPTIONS | MPD_IDLE_PARTITION         |
        MPD_IDLE_STICKER | MPD_IDLE_SUBSCRIPTION | MPD_IDLE_MESSAGE,
} mpd_idle_flag;

#define LKT_BACKLOG         32
#define COMMAND_LIST_MAX    64
#define BUFFER_OUT_MAX      16
#define MPD_VERSION         "0.21.16"

/* Macros */

#define SQLITE_PREPARE(db, stmt, SQL, goto_label)                               \
    if (sqlite3_prepare_v2((sqlite3 *) db, SQL, -1, &(stmt), 0) != SQLITE_OK) { \
        LOG_ERROR("DB", "Failed to prepare statement: %s",                      \
                      sqlite3_errmsg((sqlite3 *) db));                          \
        goto goto_label;                                                        \
    }

#define SQLITE_EXEC(db, SQL, goto_label)                                        \
    if (sqlite3_exec((sqlite3 *) db, SQL, NULL, NULL, NULL) != SQLITE_OK) {     \
        LOG_ERROR("DB", "Failed to exec statement: %s",                     \
                      sqlite3_errmsg((sqlite3 *) db));                          \
        goto goto_label;                                                        \
    }

#define SQLITE_BIND_TEXT(db, stmt, pos, text, error) {              \
    if (sqlite3_bind_text(stmt, pos, text, -1, 0) != SQLITE_OK) {   \
        LOG_ERROR("DB", "Failed to bind text %s at pos %d: %s",     \
                      (const char *) text, pos,                     \
                      sqlite3_errmsg((sqlite3 *) db));              \
        goto error;                                                 \
    }}

#define SQLITE_BIND_INT(db, stmt, pos, integer, error) {            \
    if (sqlite3_bind_int(stmt, pos, integer) != SQLITE_OK) {        \
        LOG_ERROR("DB", "Failed to bind int %d at pos %d: %s",      \
                      integer, pos, sqlite3_errmsg((sqlite3 *) db));\
        goto error;                                                 \
    }}

#define SQLITE_STEP(db, stmt, code, error)                          \
    if (sqlite3_step(stmt) != code) {                               \
        LOG_ERROR("DB", "Failed to step: %s",                       \
                      sqlite3_errmsg((sqlite3 *) db));              \
        goto error;                                                 \
    }

#define SQLITE_STEP_ROW(db, stmt, error)    SQLITE_STEP(db, stmt, SQLITE_ROW, error)
#define SQLITE_STEP_DONE(db, stmt, error)   SQLITE_STEP(db, stmt, SQLITE_DONE, error)

#define SQLITE_DO_ROLLBACK(db)                                      \
    sqlite3_exec((sqlite3 *) db, "ROLLBACK TRANSACTION;\n", NULL, NULL, NULL);

#define STR_MATCH(str1, str2)           (! strcasecmp(str1, str2))
#define STR_NMATCH(str1, str2, n)       (! strncasecmp(str1, str2, n))

#define sqlite3_column_chars BRACKETS_THAT(const char *) sqlite3_column_text

#define GET_VA_COUNT(...)   PP_NARG_(__VA_ARGS__,PP_RSEQ_N())
#define PP_NARG_(...)       PP_ARG_N(__VA_ARGS__)
#define PP_ARG_N( _1, _2, _3, _4, _5, _6, _7, _8, _9,_10,   \
                  _11,_12,_13,_14,_15,_16,_17,_18,_19,_20,  \
                  _21,_22,_23,_24,_25,_26,_27,_28,_29,_30,  \
                  _31,_32,_33,_34,_35,_36,_37,_38,_39,_40,  \
                  _41,_42,_43,_44,_45,_46,_47,_48,_49,_50,  \
                  _51,_52,_53,_54,_55,_56,_57,_58,_59,_60,  \
                  _61,_62,_63,N,...) N
#define PP_RSEQ_N() 63,62,61,60,                   \
                    59,58,57,56,55,54,53,52,51,50, \
                    49,48,47,46,45,44,43,42,41,40, \
                    39,38,37,36,35,34,33,32,31,30, \
                    29,28,27,26,25,24,23,22,21,20, \
                    19,18,17,16,15,14,13,12,11,10, \
                    9,8,7,6,5,4,3,2,1,0
