#define _POSIX_C_SOURCE 200809L
#define _DEFAULT_SOURCE

#include <common/common.h>
#include <lektor/database.h>

#include <stdbool.h>
#include <sqlite3.h>
#include <stdio.h>
#include <dirent.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include <linux/limits.h>

static inline bool
__add_kara_to_update_job(volatile sqlite3 *db, size_t id)
{
    static const char *SQL = NULL;
    if (id)
        SQL = "INSERT INTO updates (job, kara_id) "
              "SELECT MAX(update_job), ? FROM misc;";
    else
        SQL = "INSERT INTO updates (job, kara_id) "
              "SELECT MAX(update_job), last_insert_rowid() FROM misc;";
    sqlite3_stmt *stmt;
    SQLITE_PREPARE(db, stmt, SQL, error);
    SQLITE_BIND_INT(db, stmt, 1, (int) id, error);
    sqlite3_step(stmt); /* Ignore result of that one */
    return true;
error:
    return false;
}

static inline bool
__database_add_kara(volatile sqlite3 *db, const char *filename)
{
    RETURN_UNLESS(db || filename, "Invalid argument", false);
    static const char *SQL_STMT =
        "INSERT INTO "
        "kara (song_name, source_name, category, song_type, language, "
        "file_path, is_new, author_name, song_number)"
        "SELECT ?, ?, ?, ?, ?, ?, ?, ?, ?";
    static const char *SQL_STMT_WITH_ID =
        "INSERT INTO "
        "kara (song_name, source_name, category, song_type, language, "
        "file_path, is_new, author_name, song_number, id)"
        "SELECT ?, ?, ?, ?, ?, ?, ?, ?, ?, ?";
    sqlite3_stmt *stmt = NULL;
    long status = false, kara_id = 0;
    struct kara_metadata data;
    char _path[PATH_MAX], *saveptr = NULL, *token, *id = NULL;
    char *path = _path;

    /* Metadata */
    if (kara_metadata_read(&data, filename) != 0) {
        LOG_ERROR("UPDATE", "Failed to get mdt for the kara '%s'", filename);
        return false;
    }

    SQLITE_EXEC(db, "BEGIN TRANSACTION;", error);

    /* Try to find an id in the filename, reuse of variables, it's ugly */
    path[PATH_MAX - 1] = '\0';
    memcpy(path, filename, (strlen(filename) + 1) * sizeof(char));
    while (NULL != (token = strtok_r(path, "/", &saveptr))) {
        id = token;
        path = NULL;
    }
    size_t id_len = strspn(id, "0123456789");
    id[id_len] = '\0';  /* Replace the . of .mkv */
    STRTOL(kara_id, id, token, id_len);
    if (!id_len && STR_MATCH(&token[1], "mkv")) {
        /* Use the found id, most of the time, normally */
        LOG_INFO("DB", "Found id '%ld' in kara '%s'", kara_id, filename);
        SQLITE_PREPARE(db, stmt, SQL_STMT_WITH_ID, error);
        SQLITE_BIND_INT(db, stmt, 10, (int) kara_id, error);
    } else {
        /* Generate a new id */
        SQLITE_PREPARE(db, stmt, SQL_STMT, error);
    }

    if ((sqlite3_bind_text(stmt, 1, data.song_name,   -1, 0) != SQLITE_OK)  ||
        (sqlite3_bind_text(stmt, 2, data.source_name, -1, 0) != SQLITE_OK)  ||
        (sqlite3_bind_text(stmt, 3, data.category,    -1, 0) != SQLITE_OK)  ||
        (sqlite3_bind_text(stmt, 4, data.song_type,   -1, 0) != SQLITE_OK)  ||
        (sqlite3_bind_text(stmt, 5, data.language,    -1, 0) != SQLITE_OK)  ||
        (sqlite3_bind_text(stmt, 6, filename,         -1, 0) != SQLITE_OK)  ||
        (sqlite3_bind_int (stmt, 7, 0) /* TODO */            != SQLITE_OK)  ||
        (sqlite3_bind_text(stmt, 8, data.author_name, -1, 0) != SQLITE_OK)  ||
        (sqlite3_bind_int (stmt, 9, data.song_number)        != SQLITE_OK)
       ) {
        LOG_ERROR("DB", "Failed to bind for kara %s: %s", filename,
                  sqlite3_errmsg((sqlite3 *) db));
        goto error;
    }
    SQLITE_STEP_DONE(db, stmt, error);
    sqlite3_finalize(stmt);

    if (__add_kara_to_update_job(db, kara_id)) {
        SQLITE_EXEC(db, "COMMIT;", error);
        return true;
    }
error:
    sqlite3_finalize(stmt);
    SQLITE_DO_ROLLBACK(db);
    return status;
}

bool
database_update_set_available(volatile sqlite3 *db, int id)
{
    static const char *SQL_STMT =
        "UPDATE kara SET available = 1 WHERE id = ?;";
    sqlite3_stmt *stmt;
    SQLITE_PREPARE(db, stmt, SQL_STMT, error);
    SQLITE_BIND_INT(db, stmt, 1, id, error);
    if (sqlite3_step(stmt) != SQLITE_DONE)
        goto error;
    sqlite3_finalize(stmt);
    return true;
error:
    sqlite3_finalize(stmt);
    LOG_ERROR("DB", "Failed to set kara %d available", id);
    return false;
}

bool
database_update_add(volatile sqlite3 *db, const char *kara_path,
                    struct kara_metadata *mdt, uint64_t id, bool avail)
{
    RETURN_UNLESS(db && kara_path && mdt && id, "Invalid argument", false);
    static const char *SQL_STMT =
        "INSERT OR REPLACE INTO kara (song_name, source_name, category, "
        "song_type, language, file_path, is_new, author_name, "
        "song_number, id, available)"
        "SELECT ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?";
    sqlite3_stmt *stmt = NULL;
    int ret = false;

    /* From here we initialize the sqlite stmt. */

    SQLITE_PREPARE(db, stmt, SQL_STMT, error);

    if ((sqlite3_bind_text(stmt,  1, mdt->song_name,   -1, 0) != SQLITE_OK) ||
        (sqlite3_bind_text(stmt,  2, mdt->source_name, -1, 0) != SQLITE_OK) ||
        (sqlite3_bind_text(stmt,  3, mdt->category,    -1, 0) != SQLITE_OK) ||
        (sqlite3_bind_text(stmt,  4, mdt->song_type,   -1, 0) != SQLITE_OK) ||
        (sqlite3_bind_text(stmt,  5, mdt->language,    -1, 0) != SQLITE_OK) ||
        (sqlite3_bind_text(stmt,  6, kara_path,        -1, 0) != SQLITE_OK) ||
        (sqlite3_bind_int (stmt,  7, 0) /* TODO */            != SQLITE_OK) ||
        (sqlite3_bind_text(stmt,  8, mdt->author_name, -1, 0) != SQLITE_OK) ||
        (sqlite3_bind_int (stmt,  9, mdt->song_number)        != SQLITE_OK) ||
        (sqlite3_bind_int (stmt, 10, id)                      != SQLITE_OK) ||
        (sqlite3_bind_int (stmt, 11, avail)                   != SQLITE_OK)
       ) {
        LOG_ERROR("DB", "Failed to bind argument for kara %s: %s", kara_path, sqlite3_errmsg((sqlite3 *) db));
        goto error;
    }

    SQLITE_STEP_DONE(db, stmt, error);
    ret = __add_kara_to_update_job(db, id);
error:
    sqlite3_finalize(stmt);
    return ret;
}

bool
database_update(volatile sqlite3 *db, const char *kara_dir, int check_timestamp)
{
    DIR *d;
    struct dirent *dir;
    char path[PATH_MAX];
    long db_timestamp = 0;
    memset(path, 0, PATH_MAX * sizeof(char));

    if (!(d = opendir(kara_dir))) {
        LOG_ERROR("DB", "Failed to open directory '%s': %s", kara_dir,
                  strerror(errno));
        return false;
    }

    while (NULL != (dir = readdir(d))) {
        strncpy(path, kara_dir, PATH_MAX - 1);
        strncat(path, "/", PATH_MAX - 1);
        strncat(path, dir->d_name, PATH_MAX - 1);

        if (dir->d_type == DT_REG) {
            database_get_update(db, &db_timestamp, NULL, NULL);
            LOG_INFO("DB", "File is '%s'", path);
            if (check_timestamp && get_mtime(path) < db_timestamp) {
                LOG_INFO("DB", "Skip update of kara '%s' be cause of "
                         "timestamps", path);
                continue;
            }
            if (!__database_add_kara(db, path))
                LOG_WARN("DB", "Failed to add kara with path '%s' to db",
                         path);
            database_stamp(db);
        }

        else if (dir->d_type == DT_DIR          &&
                 !STR_MATCH(dir->d_name, ".")   &&
                 !STR_MATCH(dir->d_name, ".."))
            database_update(db, path, check_timestamp);
    }

    LOG_INFO("UPADTE", "Passed directory '%s'", kara_dir);
    database_updated(db);
    closedir(d);
    return true;
}

void
database_deleted_kara(volatile sqlite3 *db, int **kara_id, size_t *len)
{
    static const char *SQL =
        "SELECT kara.id, file_path FROM kara WHERE kara.id NOT IN"
        "(SELECT kara_id FROM updates JOIN misc ON job = update_job);";
    sqlite3_stmt *stmt = NULL;
    *kara_id = (int *) calloc(LKT_DEFAULT_LIST_SIZE, sizeof(int));
    *len = 0;
    RETURN_UNLESS(*kara_id, "Out of memory", NOTHING);
    SQLITE_PREPARE(db, stmt, SQL, out);
    size_t size = LKT_DEFAULT_LIST_SIZE;
    void *new;

    while (sqlite3_step(stmt) == SQLITE_ROW) {
        if (size == *len) {
            new = realloc(*kara_id, size * 2 * sizeof(int));
            GOTO_UNLESS(new, "Out of memory", out);
            size *= 2;
            *kara_id = new;
        }
        (*kara_id)[(*len)++] = sqlite3_column_int(stmt, 0);
    }

out:
    return;
}

void
database_get_update(volatile sqlite3 *db, long *timestamp,
                    long *job, int *current)
{
    static const char *SQL =
        "SELECT last_update, update_job, last_update > last_end_update "
        "FROM misc WHERE id = 42;";
    sqlite3_stmt *stmt;
    SQLITE_PREPARE(db, stmt, SQL, error);
    SQLITE_STEP_ROW(db, stmt, error);
    if (timestamp)
        *timestamp = sqlite3_column_int(stmt, 0);
    if (job)
        *job = sqlite3_column_int(stmt, 1);
    if (current)
        *current = sqlite3_column_int(stmt, 2);
    return;
error:
    LOG_WARN("DB", "Failed to get informations about the last update: %s",
             sqlite3_errmsg((sqlite3 *) db));
}

void
database_update_touch(volatile sqlite3 *db, int id)
{
    __add_kara_to_update_job(db, id);
}

void
database_update_del(volatile sqlite3 *db, int id)
{
    static const char *SQL = "DELETE FROM kara WHERE id = ?;";
    sqlite3_stmt *stmt = NULL;
    SQLITE_PREPARE(db, stmt, SQL, error);
    SQLITE_BIND_INT(db, stmt, 1, id, error);
    SQLITE_STEP_DONE(db, stmt, error);
    LOG_WARN("DB", "Deleted kara with id '%d' from database", id);
error:
    return;
}

#define sqlite_just_exec(func, query)   \
    void func (volatile sqlite3 *db) {  \
        SQLITE_EXEC(db, query, error);  \
        error: return;                  \
    }
sqlite_just_exec(database_stamp, "UPDATE misc SET last_update = strftime('%s','now');")
sqlite_just_exec(database_updated, "UPDATE misc SET last_end_update = strftime('%s','now'), update_job = update_job + 1;")
#undef sqlite_just_exec
